package com.atl.smartmaintenance.faultdetection;

import com.atl.afaultdetection.utils.InfluxDBHandler;
import com.atl.afaultdetection.utils.Measurement;
import com.atl.mcod.utils.Constants;
import java.io.Console;
import java.time.Instant;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.TreeMap;
import org.influxdb.dto.QueryResult;
import org.influxdb.dto.QueryResult.Series;

/**
 *
 * @author thanasis
 */
public class FaultDetectionManager {

    /**
     * @param args the command line arguments
     */
    
    public void detectOutliers(Map<String,Measurement> measMap) throws InterruptedException {
        
        FD fd = null;
        FDKnowledge fdKnowledge = null;
        boolean stamperPositionExplicitlyAdded = false; //MCOD with Knowledge needs stamper position to define if stamper is locked
        boolean productNumberExplicitlyAdded = false; //MCOD with Knowledge needs product number as for some products (like 3 and 12) the stamper is not clear where it is locked
        //Connect to the Influx DB creating a new InfluxDB handler
        InfluxDBHandler infDBH = null;
        String password = "";
        if(Constants.username != null){
            if(Constants.password == null){
                Scanner scanner = null;
                Console console = System.console();
                if(console == null){
                    scanner = new Scanner(System.in);
                    System.out.println("\nInfluxDB Password for "+Constants.username+": ");
                    password = scanner.next();
                } else {
                    password = new String(console.readPassword("\nInfluxDB Password for "+Constants.username+": "));
                }
            } else {
                password = Constants.password;
            }
            
        }
        infDBH = new InfluxDBHandler(Constants.host, Constants.port, Constants.database, Constants.resutlsDB, Constants.username, password, Constants.ssl);
        

        if(Constants.algorithms.contains("mcodKnowledge")){
            //stamperPosition and productNumber are needed in the MCOD with knowledge algorithm, hence if not provided using the --meas flag the are added here
            fdKnowledge = new FDKnowledge(measMap, infDBH, Constants.R_k, Constants.k_k, Constants.W, Constants.slide, Constants.outlierLife_k);
            if(!measMap.containsKey("stamperPosition")){
                measMap.put(Constants.measShortNameMap.get("stamperPosition"), new Measurement("\""+Constants.measShortNameMap.get("stamperPosition")+"\""));
                stamperPositionExplicitlyAdded = true;
            }
            if(!measMap.containsKey("productNumber")){
                measMap.put(Constants.measShortNameMap.get("productNumber"), new Measurement("\""+Constants.measShortNameMap.get("productNumber")+"\""));
                productNumberExplicitlyAdded = true;
            }
        }
        
        if(Constants.algorithms.contains("mcod")){
            fd = new FD(measMap, infDBH, Constants.R, Constants.k, Constants.W, Constants.slide, Constants.outlierLife, stamperPositionExplicitlyAdded, productNumberExplicitlyAdded);
        }
        
        //instatiate the minimum and maximum values for each measurement needed for the min-max normalization process
        for(String measName : measMap.keySet()){
            boolean excludeZeroValues = measName.contains(Constants.measShortNameMap.get("oilLevel")) || measName.contains(Constants.measShortNameMap.get("stamperPosition")) || measName.contains(Constants.measShortNameMap.get("1_max"))
                    || measName.contains(Constants.measShortNameMap.get("2_max"))  || measName.contains(Constants.measShortNameMap.get("3_max"))  || measName.contains(Constants.measShortNameMap.get("4_max"))
                    || measName.contains(Constants.measShortNameMap.get("5_max"))  || measName.contains(Constants.measShortNameMap.get("6_max"))
                    || measName.contains(Constants.measShortNameMap.get("in_tmp"))  || measName.contains(Constants.measShortNameMap.get("thick"))
                    || measName.contains(Constants.measShortNameMap.get("out_tmp"));
            boolean excludeNegativeValues = true; // we don't want negative values in any of the measurements
            measMap.get(measName).setMinValue(infDBH.getMinValue(measMap.get(measName).getMeasurementDBName(),excludeZeroValues, excludeNegativeValues));
            measMap.get(measName).setMaxValue(infDBH.getMaxValue(measMap.get(measName).getMeasurementDBName()));
            if(Constants.debug){
                System.out.println(measMap.get(measName).getMinValue());
                System.out.println(measMap.get(measName).getMaxValue());   
            }
        }
        
        //The offset used to fetch measurments from the database. It is instatiated to the slide size
        long secs = Constants.slide;
        

        
        //form the query concatanating all the participating measurments
        StringBuilder strb = new StringBuilder();
        int i = 0;
        for(String measName : measMap.keySet()){
            strb.append(measMap.get(measName).getMeasurementDBName());
            if(i < measMap.size()-1){
                strb.append(",");
            }
            i++;
        }
            
        while(true){
            QueryResult measurements = null;
            if(Constants.debug){
                measurements = infDBH.getMeasurementsDebug(strb.toString(), secs, true);
            } else {
                measurements = infDBH.getMeasurements(strb.toString(), secs, false);
            }
            //influx DB can't syncronize by time measurments from different tables.
            //rawFetchedMeasurements is used to transform the returned results to an easy to process form
            Map<String,Map<Instant,Double>> rawFetchedMeasurements = new HashMap<>();
            //processedFetchedMeasurements holds the final synced measurements
            Map<String,Map<Instant,Double>> processedFetchedMeasurements = new HashMap<>();
            
            if(!measurements.hasError() && measurements.getResults().get(0).getSeries() != null){
                //Store all the value iterators to a HashMap
                
                for(Series series : measurements.getResults().get(0).getSeries()){
                    for(List pair : series.getValues()){
                        if(!rawFetchedMeasurements.containsKey(series.getName())){
                            rawFetchedMeasurements.put(series.getName(), new TreeMap()); // use TreeMap to keep the measurements sorted by their time
                        }
                        if(pair.get(1) != null){
                            //the first item of the pair is always the date (transformed into Instant), the second one is the value (transformed into double)
                            rawFetchedMeasurements.get(series.getName()).put(Instant.parse(pair.get(0).toString()), (Double) pair.get(1)); 
                        }
                    }
                }
                
                //filter the measurements based on the dates of the smallest series.
                String smallerSeriesName = findSmallestMap(rawFetchedMeasurements);
                for(String seriesName : rawFetchedMeasurements.keySet()){
                    if(!seriesName.equals(smallerSeriesName)){
                        for(Instant key : rawFetchedMeasurements.get(smallerSeriesName).keySet()){
                            if(rawFetchedMeasurements.get(seriesName).containsKey(key)){
                                if(!processedFetchedMeasurements.containsKey(seriesName)){
                                    processedFetchedMeasurements.put(seriesName, new TreeMap<>());
                                }
                                processedFetchedMeasurements.get(seriesName).put(key, rawFetchedMeasurements.get(seriesName).get(key));
                            }
                        } 
                    }
                }
                
                if(!processedFetchedMeasurements.isEmpty() || rawFetchedMeasurements.size() == 1){
                    processedFetchedMeasurements.put(smallerSeriesName, rawFetchedMeasurements.get(smallerSeriesName));
                    rawFetchedMeasurements.clear();


                    Map<String,Iterator<Entry<Instant,Double>>> slideDataIteratorsMap = new HashMap<>();
                    Map<String,Iterator<Entry<Instant,Double>>> slideDataIteratorsMapKnowledge = new HashMap<>();
                    for(String seriesName : processedFetchedMeasurements.keySet()){
                        slideDataIteratorsMap.put(seriesName,processedFetchedMeasurements.get(seriesName).entrySet().iterator());
                        slideDataIteratorsMapKnowledge.put(seriesName,processedFetchedMeasurements.get(seriesName).entrySet().iterator());
                    }
                    if(Constants.algorithms.contains("mcod")){
                        fd.mcodNoKnowledge(slideDataIteratorsMap);
                    }
                    if(Constants.algorithms.contains("mcodKnowledge")){
                        fdKnowledge.mcodWithKnowledge(slideDataIteratorsMapKnowledge);
                    }
                }
            }
            if(Constants.debug){
                //Increment the secs for the query by a slide
                secs += Constants.slide;
//                if(secs > 6000000){
                if(secs > 2332800){
                    System.out.println(Constants.resutlsDBTable + " finished");
                    break;
                }
            } 
            if(!Constants.debug){
                Thread.sleep(Constants.slide*1000);
            }
        }
    }

    private String findSmallestMap(Map<String, Map<Instant, Double>> rawFetchedMeasurments) {
        String smallestSeriesName = "";
        int min = Integer.MAX_VALUE;
        for(String seriesName : rawFetchedMeasurments.keySet()){
            if(rawFetchedMeasurments.get(seriesName).size() < min){
                smallestSeriesName = seriesName;
                min = rawFetchedMeasurments.get(seriesName).size();
            }
        }
        return smallestSeriesName;
    }
}
